import contextlib

try:
    from asgiref.local import Local
except ImportError:
    from threading import local as Local  # noqa: N812
_thread_locals = Local()


@contextlib.contextmanager
def set_request(request):
    """
    Context processor that sets the request on the current instance
    """
    _thread_locals._request = request
    yield


def get_request():
    """
    Retrieve request from current instance
    """
    return getattr(_thread_locals, "_request", None)



import json
from copy import copy

from django.contrib.sites.models import Site
from django.core.exceptions import ImproperlyConfigured
from django.test import RequestFactory, TestCase, modify_settings, override_settings

from meta.views import Meta


@override_settings(
    META_SITE_PROTOCOL="",
    META_USE_SITES=False,
    META_USE_OG_PROPERTIES=False,
    META_USE_TWITTER_PROPERTIES=False,
    META_USE_SCHEMAORG_PROPERTIES=False,
    META_USE_JSON_LD_SCHEMA=False,
)
class MetaObjectTestCase(TestCase):
    def test_defaults(self):
        m = Meta()
        self.assertEqual(m.title, None)
        self.assertEqual(m.og_title, None)
        self.assertEqual(m.schemaorg_title, None)
        self.assertEqual(m.schemaorg_description, None)
        self.assertEqual(m.twitter_title, None)
        self.assertEqual(m.description, None)
        self.assertEqual(m.extra_props, None)
        self.assertEqual(m.extra_custom_props, None)
        self.assertEqual(m.custom_namespace, None)
        self.assertEqual(m.keywords, [])
        self.assertEqual(m.url, None)
        self.assertEqual(m.image, None)
        self.assertEqual(m.image_object, None)
        self.assertEqual(m.object_type, None)
        self.assertEqual(m.site_name, None)
        self.assertEqual(m.twitter_site, None)
        self.assertEqual(m.twitter_creator, None)
        self.assertEqual(m.locale, None)
        self.assertEqual(m.facebook_app_id, None)
        self.assertEqual(m.use_og, False)
        self.assertEqual(m.use_sites, False)
        self.assertEqual(m.use_twitter, False)
        self.assertEqual(m.use_facebook, False)
        self.assertEqual(m.use_schemaorg, False)
        self.assertEqual(m.use_json_ld, False)
        self.assertEqual(m.fb_pages, "")
        self.assertEqual(m.og_app_id, "")
        self.assertEqual(m.use_title_tag, False)
        self.assertEqual(m.schema, {"@type": m.schemaorg_type})

    def test_set_keywords(self):
        m = Meta(keywords=["foo", "bar"])
        self.assertEqual(m.keywords[0], "foo")
        self.assertEqual(m.keywords[1], "bar")

    @override_settings(META_USE_SITES=True, META_SITE_PROTOCOL="http", SITE_ID=None)
    def test_set_image_request(self):
        factory = RequestFactory()
        request = factory.get("/")
        Site.objects.create(domain=request.get_host())
        m = Meta(
            request=request,
            title="A page title",
            image="/static/image.png",
        )
        self.assertEqual(m.image, "http://testserver/static/image.png")

    @override_settings(META_INCLUDE_KEYWORDS=["baz"])
    def test_set_keywords_with_include(self):
        m = Meta(keywords=["foo", "bar"])
        self.assertEqual(m.keywords[0], "foo")
        self.assertEqual(m.keywords[1], "bar")
        self.assertEqual(m.keywords[2], "baz")

    def test_set_keywords_no_duplicate(self):
        m = Meta(keywords=["foo", "foo", "foo"])
        self.assertEqual(m.keywords[0], "foo")
        self.assertEqual(len(m.keywords), 1)

    @override_settings(META_USE_JSON_LD_SCHEMA=True)
    def test_use_json_ld(self):
        m = Meta()
        self.assertEqual(m.use_json_ld, True)

    @override_settings(META_FB_APPID="appid", META_FB_PAGES="fbpages")
    def test_pages_appid(self):
        m = Meta()
        self.assertEqual(m.fb_pages, "fbpages")
        self.assertEqual(m.og_app_id, "appid")

    @override_settings(META_DEFAULT_KEYWORDS=["foo", "bar"])
    def test_set_keywords_with_defaults(self):
        m = Meta()
        self.assertEqual(m.keywords[0], "foo")
        self.assertEqual(m.keywords[1], "bar")

    @override_settings(META_OG_NAMESPACES=["foo", "bar"])
    def test_set_namespaces(self):
        m = Meta()
        self.assertEqual(m.custom_namespace[0], "foo")
        self.assertEqual(m.custom_namespace[1], "bar")

    def test_get_full_url_with_none(self):
        m = Meta()
        self.assertEqual(m.get_full_url(None), None)

    def test_get_full_url_with_full_url(self):
        m = Meta()
        self.assertEqual(m.get_full_url("http://example.com/foo"), "http://example.com/foo")

    def test_get_full_url_without_protocol_will_raise(self):
        m = Meta()
        with self.assertRaises(ImproperlyConfigured):
            m.get_full_url("foo/bar")

    @override_settings(SITE_ID=None, META_USE_SITES=True, META_SITE_PROTOCOL="http")
    def test_get_full_url_without_site_id_will_raise(self):
        m = Meta()
        with self.assertRaises(ImproperlyConfigured):
            m.get_full_url("foo/bar")

    @override_settings(SITE_ID=None, META_USE_SITES=True, META_SITE_PROTOCOL="http")
    def test_get_full_url_without_site_id_with_request_will_not_raise(self):
        factory = RequestFactory()
        request = factory.get("/")
        Site.objects.create(domain=request.get_host())
        m = Meta(request=request)
        self.assertEqual(m.get_full_url("foo/bar"), "http://testserver/foo/bar")

    @override_settings(SITE_ID=None, META_USE_SITES=True, META_SITE_PROTOCOL="http")
    def test_get_full_url_with_fdqn_original_url(self):
        factory = RequestFactory()
        request = factory.get("/")
        Site.objects.create(domain=request.get_host())
        m = Meta(request=request)
        self.assertEqual(m.get_full_url("https://example.com/foo/bar"), "https://example.com/foo/bar")

    def test_get_full_url_without_protocol_without_schema_will_raise(self):
        m = Meta()
        with self.assertRaises(ImproperlyConfigured):
            m.get_full_url("//foo.com/foo/bar")

    @override_settings(META_SITE_PROTOCOL="http")
    def test_get_full_url_without_domain_will_raise(self):
        m = Meta()
        with self.assertRaises(ImproperlyConfigured):
            m.get_full_url("foo/bar")

    @override_settings(META_SITE_PROTOCOL="https", META_SITE_DOMAIN="foo.com")
    def test_get_full_url_with_domain_and_protocol(self):
        m = Meta()
        self.assertEqual(m.get_full_url("foo/bar"), "https://foo.com/foo/bar")

    @override_settings(META_SITE_PROTOCOL="https", META_SITE_DOMAIN="foo.com")
    def test_get_full_url_without_schema(self):
        m = Meta()
        self.assertEqual(m.get_full_url("//foo.com/foo/bar"), "https://foo.com/foo/bar")

    @override_settings(META_SITE_PROTOCOL="https", META_SITE_DOMAIN="foo.com")
    def test_get_full_url_with_absolute_path(self):
        m = Meta()
        self.assertEqual(m.get_full_url("/foo/bar"), "https://foo.com/foo/bar")

    @override_settings(META_SITE_PROTOCOL="http", META_SITE_DOMAIN="http://foo.com")
    def test_get_full_url_with_wrong_domain(self):
        m = Meta()
        self.assertEqual(m.get_full_url("/foo/bar"), "http://foo.com/foo/bar")

    @override_settings(META_SITE_PROTOCOL="https", META_SITE_DOMAIN="foo.com")
    def test_set_url(self):
        m = Meta(url="foo/bar")
        self.assertEqual(m.url, "https://foo.com/foo/bar")

    @override_settings(META_SITE_PROTOCOL="https")
    def test_set_image_object_with_full_url(self):
        media = {
            "url": "http://meta.example.com/image.gif",
            "type": "some/mime",
            "width": 100,
            "height": 100,
            "alt": "a media",
        }
        secure_media = copy(media)
        secure_media["secure_url"] = "https://meta.example.com/image.gif"
        m = Meta(image_object=media)
        self.assertEqual(m.image, "http://meta.example.com/image.gif")
        self.assertEqual(m.image_object, secure_media)

    @override_settings(META_SITE_PROTOCOL="https")
    def test_set_image_object_with_custom_secure(self):
        media = {
            "url": "http://meta.example.com/image.gif",
            "secure_url": "https://meta.example.com/custom.gif",
            "type": "some/mime",
            "width": 100,
            "height": 100,
            "alt": "a media",
        }
        m = Meta(image_object=media)
        self.assertEqual(m.image, "http://meta.example.com/image.gif")
        self.assertEqual(m.image_object, media)

    def test_set_image_with_full_url(self):
        m = Meta(image="http://meta.example.com/image.gif")
        self.assertEqual(m.image, "http://meta.example.com/image.gif")

    @override_settings(META_SITE_PROTOCOL="https", META_SITE_DOMAIN="foo.com")
    def test_set_image_object_with_absolute_path(self):
        media = {
            "url": "/img/image.gif",
            "type": "some/mime",
            "width": 100,
            "height": 100,
            "alt": "a media",
        }
        secure_media = copy(media)
        secure_media["secure_url"] = "https://foo.com/img/image.gif"
        secure_media["url"] = "https://foo.com/img/image.gif"
        m = Meta(image_object=media)
        self.assertEqual(m.image, "https://foo.com/img/image.gif")
        self.assertEqual(m.image_object, secure_media)

    @override_settings(META_SITE_PROTOCOL="https", META_SITE_DOMAIN="foo.com")
    def test_set_image_with_absolute_path(self):
        m = Meta(image="/img/image.gif")
        self.assertEqual(m.image, "https://foo.com/img/image.gif")

    @override_settings(META_SITE_PROTOCOL="https", META_SITE_DOMAIN="foo.com")
    def test_set_image_with_relative_path(self):
        m = Meta(image="img/image.gif")
        self.assertEqual(m.image, "https://foo.com/static/img/image.gif")

    @override_settings(META_SITE_PROTOCOL="https", META_SITE_DOMAIN="foo.com", META_IMAGE_URL="/thumb/")
    def test_set_image_with_image_url(self):
        m = Meta(image="img/image.gif")
        self.assertEqual(m.image, "https://foo.com/thumb/img/image.gif")

    @override_settings(
        META_SITE_PROTOCOL="https",
        META_SITE_DOMAIN="foo.com",
        META_IMAGE_URL="/thumb/",
        META_DEFAULT_IMAGE="img/image.gif",
    )
    def test_set_image_with_default_image_url(self):
        m = Meta()
        self.assertEqual(m.image, "https://foo.com/thumb/img/image.gif")

    @override_settings(META_SITE_PROTOCOL="https", META_SITE_DOMAIN="foo.com", META_DEFAULT_IMAGE="img/image.gif")
    def test_set_image_with_defaults(self):
        m = Meta()
        self.assertEqual(m.image, "https://foo.com/static/img/image.gif")

    def test_schema_org(self):
        m = Meta(schema={"foo": "bar", "list": [{"fuu": "baz", "test": "schema"}]})
        self.assertEqual(
            m.schema,
            {"foo": "bar", "list": [{"fuu": "baz", "test": "schema"}], "@type": m.schemaorg_type},
        )

    def test_as_json_ld(self):
        m = Meta(schema={"foo": "bar", "list": [{"fuu": "baz", "test": "schema"}]})
        data = m.schema
        data["@context"] = "http://schema.org"
        self.assertEqual(
            m.as_json_ld(),
            json.dumps(data),
        )

    @override_settings(META_SITE_DOMAIN="example-no-sites.com", META_SITE_PROTOCOL="http")
    @modify_settings(INSTALLED_APPS={"remove": "django.contrib.sites"})
    def test_get_full_url_without_sites(self):
        m = Meta()
        self.assertEqual(m.get_full_url("foo/bar"), "http://example-no-sites.com/foo/bar")

    @override_settings(
        META_SITE_DOMAIN="example-no-sites.com",
        META_SITE_PROTOCOL="http",
        META_USE_SITES=True,
    )
    @modify_settings(INSTALLED_APPS={"remove": "django.contrib.sites"})
    def test_get_full_url_without_sites_wrong_setting(self):
        m = Meta()
        with self.assertRaises(ImproperlyConfigured):
            self.assertEqual(m.get_full_url("foo/bar"), "http://example-no-sites.com/foo/bar")



from datetime import timedelta

import django
import pytest
from django.utils.text import slugify
from django.utils.timezone import now

from tests.example_app.models import Post

try:
    from asgiref.sync import sync_to_async
    from django.test import AsyncRequestFactory, override_settings
except ImportError:
    # stub to avoid decorator failures
    def sync_to_async(__):
        return

    pytestmark = pytest.mark.skip("asgiref not installed, skipping async tests")

minversion = pytest.mark.skipif(django.VERSION < (3, 1), reason="at least Django 3.1 required")


@sync_to_async
def get_post(title):
    post, __ = Post.objects.get_or_create(
        title=title,
        og_title="og {title}".format(title=title),
        twitter_title="twitter {title}".format(title=title),
        schemaorg_title="schemaorg {title}".format(title=title),
        slug=slugify(title),
        abstract="post abstract",
        meta_description="post meta",
        meta_keywords="post keyword1,post keyword 2",
        date_published_end=now() + timedelta(days=2),
        text="post text",
        image_url="/path/to/image",
    )
    return post


@sync_to_async
def delete_post(post):
    post.delete()


@sync_to_async
def get_meta(post, request=None):
    return post.as_meta(request)


@minversion
@pytest.mark.asyncio
@pytest.mark.django_db
async def test_mixin_on_asgi():
    with override_settings(META_USE_SITES=True, META_SITE_PROTOCOL="http", META_USE_OG_PROPERTIES=True):
        post = await get_post("first post")
        meta = await get_meta(post)
        assert meta.title == "first post"
        assert meta.og_title == "og first post"


@minversion
@pytest.mark.asyncio
@pytest.mark.django_db
async def test_mixin_on_asgi_request():
    with override_settings(META_USE_SITES=True, META_SITE_PROTOCOL="http", META_USE_OG_PROPERTIES=True):
        request = AsyncRequestFactory().get("/")
        post = await get_post("first post")
        meta = await get_meta(post, request)
        assert meta.title == "first post"
        assert meta.og_title == "og first post"



import warnings
from collections import OrderedDict

from django import template
from django.apps import apps
from django.utils.html import escape
from django.utils.safestring import mark_safe

from ..settings import get_setting

register = template.Library()

# Use sekizai if installed, otherwise define a templatetag stub
if apps.is_installed("sekizai"):
    from sekizai.templatetags.sekizai_tags import Addtoblock

    register.tag("addtoblock", Addtoblock)
else:
    from meta.compat import addtoblock

    register.tag("addtoblock", addtoblock)


@register.simple_tag
def meta(name, content):
    """
    Generates a meta tag according to the following markup:

    <meta name="{name}" content="{content}">

    :param name: meta name
    :param content: content value
    """
    return custom_meta("name", name, content)


@register.simple_tag
def custom_meta(attr, name, content):
    """
    Generates a custom meta tag:

    <meta {attr}="{name}" content="{content}">

    :param attr: meta attribute name
    :param name: meta name
    :param content: content value
    """
    return '<meta {attr}="{name}" content="{content}">'.format(
        attr=escape(attr), name=escape(name), content=escape(content)
    )


@register.simple_tag
def meta_list(name, lst):
    """
    Renders in a single meta a list of values (e.g.: keywords list)

    :param name: meta name
    :param lst: values
    """
    try:
        return custom_meta("name", name, ", ".join(lst))
    except Exception:
        return ""


@register.simple_tag
def meta_extras(extra_props):
    """
    Generates the markup for a list of meta tags

    Each key,value paur is passed to :py:func:meta to generate the markup

    :param extra_props: dictionary of additional meta tags
    """
    if not extra_props:
        return ""
    return " ".join([meta(name, extra_props[name]) if extra_props[name] else "" for name in extra_props])


@register.simple_tag
def custom_meta_extras(extra_custom_props):
    """
    Generates the markup for a list of custom meta tags

    Each tuple is passed to :py:func:custom_meta to generate the markup

    :param extra_custom_props: list of tuple of additional meta tags
    """
    if not extra_custom_props:
        return ""
    return " ".join(
        [
            custom_meta(name_key, name_value, content) if content else ""
            for name_key, name_value, content in extra_custom_props
        ]
    )


@register.simple_tag
def title_prop(value):
    """
    Title tag

    :param value: title value
    """
    return "<title>%s</title>" % escape(value)


@register.simple_tag
def generic_prop(namespace, name, value):
    """
    Generic property setter that allows to create custom namespaced meta
    e.g.: fb:profile_id.
    """
    return custom_meta("property", "{}:{}".format(namespace, name), value)


@register.simple_tag
def og_prop(name, value):
    """
    Generic OpenGraph property

    :param name: property name (without 'og:' namespace)
    :param value: property value
    """
    if not isinstance(value, dict) and name in get_setting("OG_SECURE_URL_ITEMS") and value.startswith("https"):
        data = {name: value, "%s:secure_url" % name: value}
    elif not isinstance(value, dict):
        data = {
            name: value,
        }
    else:
        data = {"{}:{}".format(name, key): val for key, val in value.items()}
    data = OrderedDict(sorted(data.items()))
    content = [custom_meta("property", "og:%s" % key, val) for key, val in data.items()]
    return "\n".join(content)


@register.simple_tag
def facebook_prop(name, value):
    """
    Generic Facebook property

    :param name: property name (without 'fb:' namespace)
    :param value: property value
    """
    return custom_meta("property", "fb:%s" % name, value)


@register.simple_tag
def twitter_prop(name, value):
    """
    Generic Twitter property

    :param name: property name (without 'twitter:' namespace)
    :param value: property value
    """
    return custom_meta("name", "twitter:%s" % name, value)


@register.simple_tag
def schemaorg_prop(name, value):
    """
    Generic Schema.org property

    :param name: property name
    :param value: property value
    """
    return custom_meta("itemprop", name, value)


@register.simple_tag
def schemaorg_html_scope(value):
    """
    This is meant to be used as attribute to html / body or other tags to
    define schema.org type

    :param value: declared scope
    """
    return ' itemscope itemtype="https://schema.org/%s" ' % escape(value)


@register.simple_tag
def schemaorg_scope(value):
    """
    Alias for googleplus_html_scope

    :param value: declared scope
    """
    return schemaorg_html_scope(value)


@register.simple_tag
def googleplus_prop(name, value):
    """
    Legacy Google+ property
    """
    warnings.warn("googleplus_prop will be removed in version 3.0", PendingDeprecationWarning, stacklevel=2)
    return schemaorg_prop(name, value)


@register.simple_tag
def googleplus_html_scope(value):
    """
    Legacy Google+ scope
    """
    warnings.warn("googleplus_html_scope will be removed in version 3.0", PendingDeprecationWarning, stacklevel=2)
    return schemaorg_html_scope(value)


@register.simple_tag
def googleplus_scope(value):
    """
    Legacy Google+ scope
    """
    warnings.warn("googleplus_scope will be removed in version 3.0", PendingDeprecationWarning, stacklevel=2)
    return schemaorg_html_scope(value)


@register.simple_tag(takes_context=True)
def meta_namespaces(context):
    """
    Include OG namespaces. To be used in the <head> tag.
    """
    # do nothing if meta is not in context
    if not context.get("meta"):
        return ""

    meta = context["meta"]
    namespaces = ["og: http://ogp.me/ns#"]

    # add Facebook namespace
    if meta.use_facebook:
        namespaces.append("fb: http://ogp.me/ns/fb#")

    # add custom namespaces
    # needs to be after Facebook
    if meta.custom_namespace:
        custom_namespaces = meta.custom_namespace
        if isinstance(meta.custom_namespace, str):
            custom_namespaces = [meta.custom_namespace]
        for ns in custom_namespaces:
            custom_namespace = "{0}: http://ogp.me/ns/{0}#".format(ns)
            namespaces.append(custom_namespace)

    return mark_safe(' prefix="{}"'.format(" ".join(namespaces)))


@register.simple_tag(takes_context=True)
def meta_namespaces_schemaorg(context):
    """
    Include Schema.org attributes. To be used in the <html> or <body> tag.
    """
    # do nothing if meta is not in context or if G+ is not enabled
    if not context.get("meta") or not context["meta"].use_schemaorg:
        return ""
    return mark_safe(' itemscope itemtype="https://schema.org/{}" '.format(context["meta"].schemaorg_type))


@register.simple_tag(takes_context=True)
def meta_namespaces_gplus(context):
    """
    Legacy Google+ attributes.
    """
    warnings.warn("meta_namespaces_gplus will be removed in version 3.0", PendingDeprecationWarning, stacklevel=2)
    return meta_namespaces_schemaorg(context)



import warnings
from copy import copy

from django.db.models import Manager
from django.utils.functional import cached_property

from .settings import get_setting
from .utils import get_request, set_request
from .views import FullUrlMixin

NEED_REQUEST_OBJECT_ERR_MSG = (
    "Meta models needs request objects when initializing if sites framework "
    "is not used. See META_USE_SITES setting."
).strip()


class ModelMeta(FullUrlMixin):
    """
    Meta information mixin.
    """

    _metadata = {}
    """
    Metadata configuration dictionary

    `_metadata` dict values can be:

        * name of object method taking the field name as parameter
        * name of object method taking no parameters
        * name of object attribute
        * name of callable taking the field name as parameter
        * name of callable taking no parameters
        * literal value

    They are checked in the given order: the first that matches is returned.

    Callable must be available in the module (i.e.: imported if not defined in the module itself)
    """
    _metadata_default = {
        "title": False,
        "og_title": False,
        "twitter_title": False,
        "schemaorg_title": False,
        "description": False,
        "og_description": False,
        "twitter_description": False,
        "schemaorg_description": False,
        "keywords": False,
        "image": get_setting("DEFAULT_IMAGE"),
        "image_object": None,
        "image_width": False,
        "image_height": False,
        "object_type": get_setting("DEFAULT_TYPE"),
        "og_type": get_setting("FB_TYPE"),
        "og_app_id": get_setting("FB_APPID"),
        "og_profile_id": get_setting("FB_PROFILE_ID"),
        "og_publisher": get_setting("FB_PUBLISHER"),
        "og_author_url": get_setting("FB_AUTHOR_URL"),
        "fb_pages": get_setting("FB_PAGES"),
        "twitter_type": get_setting("TWITTER_TYPE"),
        "twitter_site": get_setting("TWITTER_SITE"),
        "twitter_author": get_setting("TWITTER_AUTHOR"),
        "schemaorg_type": get_setting("SCHEMAORG_TYPE"),
        "published_time": False,
        "modified_time": False,
        "expiration_time": False,
        "tag": False,
        "url": False,
        "locale": False,
        "custom_namespace": get_setting("OG_NAMESPACES"),
    }
    _schema = {}
    """
    schema.org properties dictionary

    `_metadata` dict values can be:

        * name of object method taking the field name as parameter
        * name of object method taking no parameters
        * name of object attribute
        * name of callable taking the field name as parameter
        * name of callable taking no parameters
        * literal value

    They are checked in the given order: the first that matches is returned.

    Callable must be available in the module (i.e.: imported if not defined in the module itself)

    If the resulting value is a :py:class:`~meta.models.ModelMeta` or :py:class:`~meta.views.Meta` instance
    its schema is set in the schema.org dataset.

    See :ref:`a sample implementation <schema.model>`.
    """

    def get_meta(self, request=None):
        """
        Retrieve the meta data configuration
        """
        metadata = copy(self._metadata_default)
        metadata.update(self._metadata)
        return metadata

    def _retrieve_data(self, request, metadata):
        """
        Build the data according to the metadata configuration
        """
        with set_request(request):
            for field, value in metadata.items():
                if value:
                    data = self._get_meta_value(field, value)
                    yield field, data

    def _get_meta_value(self, field, value):
        """
        Build metadata values from :py:attr:`_metadata`

        :param field: metadata field name
        :param value: provided value
        :return: data
        """

        def process_value(item):
            if isinstance(item, Manager):
                return list(item.all())
            elif callable(item):
                try:
                    return item(field)
                except TypeError:
                    return item()
            return item

        if value:
            try:
                return process_value(getattr(self, value))
            except (AttributeError, TypeError):
                return value

    def as_meta(self, request=None):
        """
        Populates the :py:class:`~meta.views.Meta` object  with values from :py:attr:`_metadata`

        :param request: optional request object. Used to build the correct URI for linked objects
        :return: Meta object
        """
        from meta.views import Meta

        metadata = self.get_meta(request)
        meta = Meta(request=request, obj=self)
        for field, data in self._retrieve_data(request, metadata):
            setattr(meta, field, data)
        for field in ("og_title", "twitter_title", "schemaorg_title"):
            generaltitle = getattr(meta, "title", False)
            if not getattr(meta, field, False) and generaltitle:
                setattr(meta, field, generaltitle)
        for field in ("og_description", "twitter_description", "schemaorg_description"):
            generaldesc = getattr(meta, "description", False)
            if not getattr(meta, field, False) and generaldesc:
                setattr(meta, field, generaldesc)
        if self._schema:
            meta.schema = self.schema
        return meta

    @cached_property
    def schema(self):
        """
        Schema.org object description

        :return: dict
        """
        schema = {}
        for field, value in self._schema.items():
            if value:
                schema[field] = self._get_meta_value(field, value)
        return schema

    def get_request(self):
        """
        Retrieve request from current instance
        """
        warnings.warn(
            "use meta.utils.get_request function, ModelMeta.get_request will be removed in version 3.0",
            PendingDeprecationWarning,
            stacklevel=2,
        )
        return get_request()

    def get_author(self):
        """
        Retrieve the author object. This is meant to be overridden in the model
        to return the actual author instance (e.g.: the user object).
        """

        class Author:
            fb_url = None
            twitter_profile = None
            schemaorg_profile = None

            def get_full_name(self):  # pragma: no cover
                return None

        return Author()

    def get_author_url(self):
        """
        Sample method to return the author facebook URL
        """
        try:
            return self.get_author().fb_url
        except AttributeError:  # pragma: no cover
            return ""

    def get_author_name(self):
        """
        Sample method to return the author full name
        """
        try:
            return self.get_author().get_full_name()
        except AttributeError:  # pragma: no cover
            return ""

    def get_author_twitter(self):
        """
        Sample method to return the author twitter account
        """
        try:
            return self.get_author().twitter_profile
        except AttributeError:  # pragma: no cover
            return ""

    def get_author_schemaorg(self):
        """
        Sample method to return the author Schema.org URL
        """
        try:
            return self.get_author().schemaorg_profile
        except AttributeError:  # pragma: no cover
            return ""

    def get_meta_protocol(self):
        """
        Current http protocol
        """
        return self.get_protocol()

    def build_absolute_uri(self, url):
        """
        Return the full url for the provided url argument
        """
        request = get_request()
        if request:
            return request.build_absolute_uri(url)

        if not get_setting("USE_SITES"):
            raise RuntimeError(NEED_REQUEST_OBJECT_ERR_MSG)

        return self._get_full_url(url)

    def mainEntityOfPage(self):
        return {"@type": "WebPage", "@id": self.build_absolute_uri(self.get_absolute_url())}

    @property
    def _local_key(self):
        return "{}:{}:{}".format(self._meta.app_label, self._meta.model_name, self.pk)



import warnings
from copy import copy

from django.test import TestCase, override_settings

from meta.views import Meta, MetadataMixin


@override_settings(
    META_SITE_PROTOCOL="",
    META_USE_SITES=False,
    META_USE_OG_PROPERTIES=False,
    META_USE_TWITTER_PROPERTIES=False,
    META_USE_SCHEMAORG_PROPERTIES=False,
    META_USE_JSON_LD_SCHEMA=False,
)
class MetadataMixinTestCase(TestCase):
    def test_get_meta_class(self):
        m = MetadataMixin()
        self.assertEqual(m.get_meta_class(), Meta)

    def test_get_meta_title(self):
        m = MetadataMixin()
        self.assertEqual(m.get_meta_title(), None)

        m.title = "Foo"
        self.assertEqual(m.get_meta_title(), "Foo")

    def test_get_meta_og_title(self):
        m = MetadataMixin()
        self.assertEqual(m.get_meta_og_title(), None)

        m.og_title = "Foo"
        self.assertEqual(m.get_meta_og_title(), "Foo")

    def test_get_meta_twitter_title(self):
        m = MetadataMixin()
        self.assertEqual(m.get_meta_twitter_title(), None)

        m.twitter_title = "Foo"
        self.assertEqual(m.get_meta_twitter_title(), "Foo")

    def test_get_meta_schemaorg_title(self):
        m = MetadataMixin()
        self.assertEqual(m.get_meta_schemaorg_title(), None)

        m.schemaorg_title = "Foo"
        self.assertEqual(m.get_meta_schemaorg_title(), "Foo")

    def test_get_meta_schemaorg_description(self):
        m = MetadataMixin()
        self.assertEqual(m.get_meta_schemaorg_description(), None)

        m.schemaorg_description = "Foo"
        self.assertEqual(m.get_meta_schemaorg_description(), "Foo")

    def test_get_meta_description(self):
        m = MetadataMixin()
        self.assertEqual(m.get_meta_description(), None)

        m.description = "Foo"
        self.assertEqual(m.get_meta_description(), "Foo")

    def test_get_meta_url(self):
        m = MetadataMixin()
        self.assertEqual(m.get_meta_url(), None)

        m.url = "/foo/bar"
        self.assertEqual(m.get_meta_url(), "/foo/bar")

    def test_get_meta_image(self):
        m = MetadataMixin()
        self.assertEqual(m.get_meta_image(), None)

        m.image = "img/foo.gif"

        self.assertEqual(m.get_meta_image(), "img/foo.gif")

    def test_get_meta_image_object(self):
        m = MetadataMixin()
        self.assertEqual(m.get_meta_image(), None)
        media = {
            "url": "http://meta.example.com/image.gif",
            "secure_url": "https://meta.example.com/custom.gif",
            "type": "some/mime",
            "width": 100,
            "height": 100,
            "alt": "a media",
        }
        m.image_object = media

        self.assertEqual(m.get_meta_image_object(), media)
        self.assertEqual(m.get_meta_image(), media["url"])

    def test_get_meta_object_tye(self):
        m = MetadataMixin()
        self.assertEqual(m.get_meta_object_type(), None)

        m.object_type = "bar"
        self.assertEqual(m.get_meta_object_type(), "bar")

    @override_settings(META_SITE_DOMAIN="somedomain.com")
    def test_get_domain(self):
        m = MetadataMixin()
        self.assertEqual(m.get_domain(), "somedomain.com")

    @override_settings(META_SITE_PROTOCOL="http")
    def test_get_domain_http(self):
        m = MetadataMixin()
        self.assertEqual(m.get_protocol(), "http")

    @override_settings(META_SITE_TYPE="foo")
    def test_get_meta_object_type_with_setting(self):
        m = MetadataMixin()
        self.assertEqual(m.get_meta_object_type(), "foo")

    def test_get_meta_site_name(self):
        m = MetadataMixin()
        self.assertEqual(m.get_meta_site_name(), None)

        m.site_name = "Foo"
        self.assertEqual(m.get_meta_site_name(), "Foo")

    @override_settings(META_SITE_PROTOCOL="http", META_SITE_DOMAIN="http://foo.com")
    def test_get_full_url_with_wrong_domain(self):
        m = MetadataMixin()
        self.assertEqual(m._get_full_url("/foo/bar"), "http://foo.com/foo/bar")

    @override_settings(META_SITE_NAME="Foo")
    def test_get_meta_site_name_with_setting(self):
        m = MetadataMixin()
        self.assertEqual(m.get_meta_site_name(), "Foo")

    def test_get_meta_extra_props(self):
        m = MetadataMixin()
        self.assertEqual(m.get_meta_extra_props(), None)

        m.extra_props = {"app_name": "Foo", "app_id": "Bar"}
        self.assertEqual(m.get_meta_extra_props(), {"app_name": "Foo", "app_id": "Bar"})

    def test_get_meta_extra_custom_props(self):
        m = MetadataMixin()
        self.assertEqual(m.get_meta_extra_custom_props(), None)

        m.extra_custom_props = [
            ("property", "app_name", "Foo"),
            ("property", "app_id", "Bar"),
        ]
        self.assertEqual(
            m.get_meta_extra_custom_props(), [("property", "app_name", "Foo"), ("property", "app_id", "Bar")]
        )

    def test_get_meta_custom_namespace(self):
        m = MetadataMixin()
        self.assertEqual(m.get_meta_custom_namespace(), None)

        m.custom_namespace = "my-website"
        self.assertEqual(m.get_meta_custom_namespace(), "my-website")

        with override_settings(META_OG_NAMESPACES=["foo", "bar"]):
            m = MetadataMixin()
            self.assertEqual(m.get_meta_custom_namespace(), ["foo", "bar"])

    def test_get_meta_twitter_site(self):
        m = MetadataMixin()
        self.assertEqual(m.get_meta_twitter_site(), None)

        m.twitter_site = "@foo"
        self.assertEqual(m.get_meta_twitter_site(), "@foo")

    def test_get_meta_twitter_creator(self):
        m = MetadataMixin()
        self.assertEqual(m.get_meta_twitter_creator(), None)

        m.twitter_creator = "@foo"
        self.assertEqual(m.get_meta_twitter_creator(), "@foo")

    def test_get_meta_twitter_card(self):
        with warnings.catch_warnings(record=True) as w:
            warnings.simplefilter("always")
            m = MetadataMixin()
            self.assertEqual(m.get_meta_twitter_card(), None)
            assert len(w) == 1
            assert issubclass(w[-1].category, PendingDeprecationWarning)

            m.twitter_card = "summary"
            assert len(w) == 2
            assert issubclass(w[-1].category, PendingDeprecationWarning)

            self.assertEqual(m.twitter_card, "summary")
            assert len(w) == 3
            assert issubclass(w[-1].category, PendingDeprecationWarning)

            self.assertEqual(m.get_meta_twitter_card(), "summary")
            assert len(w) == 4
            assert issubclass(w[-1].category, PendingDeprecationWarning)

    def test_get_meta_twitter_type(self):
        with warnings.catch_warnings(record=True) as w:
            warnings.simplefilter("always")
            m = MetadataMixin()
            self.assertEqual(m.get_meta_twitter_type(), None)
            assert len(w) == 0

            m.twitter_type = "summary"
            self.assertEqual(m.get_meta_twitter_type(), "summary")
            assert len(w) == 0

    def test_no_schema_org(self):
        m = MetadataMixin()
        self.assertEqual(m.schema, {})

    def test_get_meta_facebook_app_id(self):
        m = MetadataMixin()
        self.assertEqual(m.get_meta_facebook_app_id(), None)

        m.facebook_app_id = "12345"
        self.assertEqual(m.get_meta_facebook_app_id(), "12345")

    def test_get_meta_locale(self):
        m = MetadataMixin()
        self.assertEqual(m.get_meta_locale(), None)

        m.locale = "en_US"
        self.assertEqual(m.get_meta_locale(), "en_US")

    @override_settings(
        META_SITE_PROTOCOL="http",
        META_SITE_DOMAIN="foo.com",
        META_USE_SITES=False,
        META_FB_PAGES="fbpages",
        META_FB_APPID="appid",
    )
    def test_get_meta(self):
        media = {
            "url": "images/foo.gif",
            "type": "some/mime",
            "width": 100,
            "height": 100,
            "alt": "a media",
        }
        full_url_media = copy(media)
        full_url_media["url"] = "http://foo.com/static/images/foo.gif"
        m = MetadataMixin()
        m.title = "title"
        m.description = "description"
        m.keywords = ["foo", "bar"]
        m.url = "some/path"
        m.image_object = media

        meta_object = m.get_meta()

        self.assertTrue(type(meta_object), Meta)
        self.assertEqual(meta_object.title, "title")
        self.assertEqual(meta_object.description, "description")
        self.assertEqual(meta_object.url, "http://foo.com/some/path")
        self.assertEqual(meta_object.keywords, ["foo", "bar"])
        self.assertEqual(meta_object.image, "http://foo.com/static/images/foo.gif")
        self.assertEqual(meta_object.image_object, media)

    def test_get_context(self):
        media = {
            "url": "images/foo.gif",
            "type": "some/mime",
            "width": 100,
            "height": 100,
            "alt": "a media",
        }
        full_url_media = copy(media)
        full_url_media["url"] = "https://foo.com/static/images/foo.gif"
        full_url_media["secure_url"] = "https://foo.com/static/images/foo.gif"

        class Super:
            def get_context_data(self):
                return {}

        class View(MetadataMixin, Super):
            title = "title"
            description = "description"
            keywords = ["foo", "bar"]
            url = "some/path"
            image_object = media

        with override_settings(META_SITE_PROTOCOL="https", META_SITE_DOMAIN="foo.com", META_USE_SITES=False):
            v = View()

            context = v.get_context_data()

            self.assertTrue("meta" in context)
            self.assertTrue(type(context["meta"]), Meta)
            self.assertEqual(context["meta"].url, "https://foo.com/some/path")
            self.assertEqual(context["meta"].keywords, ["foo", "bar"])
            self.assertEqual(context["meta"].image, "https://foo.com/static/images/foo.gif")
            self.assertEqual(context["meta"].image_object, full_url_media)



import sys
import warnings

from django.test import TestCase

from meta.templatetags.meta import (
    custom_meta,
    custom_meta_extras,
    facebook_prop,
    generic_prop,
    googleplus_html_scope,
    googleplus_prop,
    googleplus_scope,
    meta,
    meta_extras,
    meta_list,
    meta_namespaces,
    meta_namespaces_gplus,
    meta_namespaces_schemaorg,
    og_prop,
    schemaorg_html_scope,
    schemaorg_prop,
    schemaorg_scope,
    title_prop,
    twitter_prop,
)
from meta.views import Meta


class OgPropTestCase(TestCase):
    def test_og_prop_basically_works(self):
        self.assertEqual(og_prop("type", "website"), '<meta property="og:type" content="website">')

    def test_og_prop_secure_url(self):
        for content in ("image", "video", "audio"):
            self.assertEqual(
                og_prop(content, "https://some/{}".format(content)),
                '<meta property="og:{0}" content="https://some/{0}">\n'
                '<meta property="og:{0}:secure_url" content="https://some/{0}">'.format(content),
            )
            self.assertEqual(
                og_prop(content, "http://some/{}".format(content)),
                '<meta property="og:{0}" content="http://some/{0}">'.format(content),
            )
        self.assertEqual(
            og_prop("random_type", "https://some/random_type"),
            '<meta property="og:random_type" content="https://some/random_type">',
        )
        self.assertEqual(
            og_prop("random_type", "http://some/random_type"),
            '<meta property="og:random_type" content="http://some/random_type">',
        )

    def test_og_full_media_data(self):
        media = {
            "url": "https://some/media.url",
            "secure_url": "https://some/media.url",
            "type": "some/mime",
            "width": 100,
            "height": 100,
            "alt": "a media",
        }
        for content in ("image", "video", "audio"):
            self.assertEqual(
                og_prop(content, media),
                '<meta property="og:{0}:alt" content="a media">\n'
                '<meta property="og:{0}:height" content="100">\n'
                '<meta property="og:{0}:secure_url" content="https://some/media.url">\n'
                '<meta property="og:{0}:type" content="some/mime">\n'
                '<meta property="og:{0}:url" content="https://some/media.url">\n'
                '<meta property="og:{0}:width" content="100">'
                "".format(content),
            )

    def test_generic_prop_basically_works(self):
        self.assertEqual(generic_prop("og", "type", "website"), '<meta property="og:type" content="website">')

    def test_generic_prop_escapes_xss(self):
        self.assertEqual(
            generic_prop("og", 't"y&p<e', "web&site"),
            '<meta property="og:t&quot;y&amp;p&lt;e" content="web&amp;site">',
        )


class MetaTestCase(TestCase):
    def test_meta_basically_works(self):
        self.assertEqual(
            meta("description", "Awesome website about ponies"),
            '<meta name="description" content="Awesome website about ponies">',
        )

    def test_meta_escapes_xss(self):
        self.assertEqual(
            meta('desc"rip&tion', "Awesome website < about ponies"),
            '<meta name="desc&quot;rip&amp;tion" content="Awesome website &lt; about ponies">',
        )


class CustomMetaTestCase(TestCase):
    def test_custom_meta_basically_works(self):
        self.assertEqual(custom_meta("property", "foo", "bar"), '<meta property="foo" content="bar">')

    def test_custom_meta_escapes_xss(self):
        self.assertEqual(
            custom_meta("prop&erty", 'fo"o', "b<ar"), '<meta prop&amp;erty="fo&quot;o" content="b&lt;ar">'
        )


class TwitterPropTestCase(TestCase):
    def test_twitter_basically_works(self):
        self.assertEqual(twitter_prop("foo", "bar"), '<meta name="twitter:foo" content="bar">')

    def test_twitter_escapes_xss(self):
        self.assertEqual(twitter_prop('fo"o', "b<ar"), '<meta name="twitter:fo&quot;o" content="b&lt;ar">')


class FacebookPropTestCase(TestCase):
    def test_facebook_basically_works(self):
        self.assertEqual(facebook_prop("foo", "bar"), '<meta property="fb:foo" content="bar">')


class GooglePlusPropTestcase(TestCase):
    def setUp(self):
        # The __warningregistry__'s need to be in a pristine state for tests
        # to work properly.
        for v in sys.modules.values():
            if getattr(v, "__warningregistry__", None):
                v.__warningregistry__ = {}

    def test_google_plus_basically_works(self):
        with warnings.catch_warnings(record=True) as w:
            warnings.simplefilter("always")
            self.assertEqual(googleplus_prop("foo", "bar"), '<meta itemprop="foo" content="bar">')
            assert len(w) == 1
            assert issubclass(w[-1].category, PendingDeprecationWarning)

    def test_google_plus_html_scope_works(self):
        with warnings.catch_warnings(record=True) as w:
            warnings.simplefilter("always")
            self.assertEqual(googleplus_html_scope("bar"), ' itemscope itemtype="https://schema.org/bar" ')
            assert len(w) == 1
            assert issubclass(w[-1].category, PendingDeprecationWarning)

    def test_google_plus_scope_works(self):
        with warnings.catch_warnings(record=True) as w:
            warnings.simplefilter("always")
            self.assertEqual(googleplus_scope("bar"), ' itemscope itemtype="https://schema.org/bar" ')
            assert len(w) == 1
            assert issubclass(w[-1].category, PendingDeprecationWarning)

    def test_google_plus_escapes_xss(self):
        with warnings.catch_warnings(record=True) as w:
            warnings.simplefilter("always")
            self.assertEqual(googleplus_prop('fo"o', "b<ar"), '<meta itemprop="fo&quot;o" content="b&lt;ar">')
            assert len(w) == 1
            assert issubclass(w[-1].category, PendingDeprecationWarning)

    def test_meta_namespaces_gplus(self):
        context_use_schemaorg = {"meta": Meta(use_schemaorg=True)}
        context_no_use_schemaorg = {"meta": Meta(use_schemaorg=False)}

        with warnings.catch_warnings(record=True) as w:
            warnings.simplefilter("always")
            self.assertEqual(meta_namespaces_gplus({}), "")
            assert len(w) == 1
            assert issubclass(w[-1].category, PendingDeprecationWarning)

        with warnings.catch_warnings(record=True) as w:
            warnings.simplefilter("always")
            self.assertEqual(meta_namespaces_gplus(context_no_use_schemaorg), "")
            assert len(w) == 1
            assert issubclass(w[-1].category, PendingDeprecationWarning)

        with warnings.catch_warnings(record=True) as w:
            warnings.simplefilter("always")
            self.assertEqual(
                meta_namespaces_gplus(context_use_schemaorg), ' itemscope itemtype="https://schema.org/Article" '
            )
            assert len(w) == 1
            assert issubclass(w[-1].category, PendingDeprecationWarning)


class SchemaOrgPropTestcase(TestCase):
    def test_schemaorg_basically_works(self):
        self.assertEqual(schemaorg_prop("foo", "bar"), '<meta itemprop="foo" content="bar">')

    def test_schemaorg_html_scope_works(self):
        self.assertEqual(schemaorg_html_scope("bar"), ' itemscope itemtype="https://schema.org/bar" ')

    def test_schemaorg_scope_works(self):
        self.assertEqual(schemaorg_scope("bar"), ' itemscope itemtype="https://schema.org/bar" ')

    def test_schemaorg_escapes_xss(self):
        self.assertEqual(schemaorg_prop('fo"o', "b<ar"), '<meta itemprop="fo&quot;o" content="b&lt;ar">')

    def test_meta_namespaces_gplus(self):
        context_use_schemaorg = {"meta": Meta(use_schemaorg=True)}
        context_no_use_schemaorg = {"meta": Meta(use_schemaorg=False)}

        self.assertEqual(meta_namespaces_schemaorg({}), "")

        self.assertEqual(meta_namespaces_schemaorg(context_no_use_schemaorg), "")

        self.assertEqual(
            meta_namespaces_schemaorg(context_use_schemaorg), ' itemscope itemtype="https://schema.org/Article" '
        )


class MetaListTestCase(TestCase):
    def test_meta_list_basically_works(self):
        self.assertEqual(
            meta_list("keywords", ["foo", "bar", "baz"]), '<meta name="keywords" content="foo, bar, baz">'
        )

    def test_meta_list_with_non_list_value(self):
        self.assertEqual(meta_list("keywords", 12), "")

    def test_meta_list_escapes_xss(self):
        self.assertEqual(
            meta_list("keywords", ['fo"o', "bar", "b<az"]), '<meta name="keywords" content="fo&quot;o, bar, b&lt;az">'
        )


class MetaExtrasTestCase(TestCase):
    def test_meta_extras_basically_works(self):
        result = meta_extras(None)
        self.assertEqual(result, "")
        result = meta_extras({"type": "foo", "image_width": "bar"})
        self.assertTrue('<meta name="type" content="foo">' in result)
        self.assertTrue('<meta name="image_width" content="bar">' in result)


class CustomMetaExtrasTestCase(TestCase):
    def test_custom_meta_extras_basically_works(self):
        result = custom_meta_extras(None)
        self.assertEqual(result, "")
        result = custom_meta_extras([("property", "type", "foo"), ("key", "image_width", "bar")])
        self.assertTrue('<meta property="type" content="foo">' in result)
        self.assertTrue('<meta key="image_width" content="bar">' in result)


class MetaNamespaceTestCase(TestCase):
    def test_meta_namespaces_no_meta_in_context(self):
        context = {}
        result = meta_namespaces(context)
        expected = ""
        self.assertEqual(result, expected)

    def test_meta_namespaces_default(self):
        context = {"meta": Meta()}
        result = meta_namespaces(context)
        expected = ' prefix="og: http://ogp.me/ns#"'
        self.assertEqual(result, expected)

    def test_meta_namespaces_facebook(self):
        context = {"meta": Meta(use_facebook=True)}
        result = meta_namespaces(context)
        expected = ' prefix="og: http://ogp.me/ns# fb: http://ogp.me/ns/fb#"'
        self.assertEqual(result, expected)

    def test_meta_namespaces_custom(self):
        context = {"meta": Meta(custom_namespace="my-website")}
        result = meta_namespaces(context)
        expected = ' prefix="og: http://ogp.me/ns# my-website: http://ogp.me/ns/my-website#"'
        self.assertEqual(result, expected)

    def test_meta_namespaces_facebook_and_custom(self):
        context = {"meta": Meta(use_facebook=True, custom_namespace="my-website")}
        result = meta_namespaces(context)
        expected = ' prefix="og: http://ogp.me/ns# fb: http://ogp.me/ns/fb# my-website: http://ogp.me/ns/my-website#"'
        self.assertEqual(result, expected)


class TitlePropTestCase(TestCase):
    def test_title_prop_basically_works(self):
        self.assertEqual(title_prop("I love django-meta app!"), "<title>I love django-meta app!</title>")

    def test_title_prop_escapes_xss(self):
        self.assertEqual(
            title_prop('I love "django-meta" app! >&:-)'),
            "<title>I love &quot;django-meta&quot; app! &gt;&amp;:-)</title>",
        )



import warnings
from datetime import timedelta

from app_helper.base_test import BaseTestCase
from django.core.exceptions import ImproperlyConfigured
from django.test.utils import override_settings
from django.utils import timezone

from meta.models import ModelMeta
from meta.settings import get_setting
from meta.templatetags.meta_extra import generic_prop, googleplus_html_scope

from .example_app.models import Comment, Post, Publisher


class TestMeta(BaseTestCase):
    post = None

    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()
        cls.publisher, __ = Publisher.objects.get_or_create(name="publisher name")
        cls.post, __ = Post.objects.get_or_create(
            title="a title",
            og_title="og title",
            twitter_title="twitter title",
            schemaorg_title="schemaorg title",
            schemaorg_description="schemaorg description",
            slug="title",
            abstract="post abstract",
            meta_description="post meta",
            meta_keywords="post keyword1,post keyword 2",
            author=cls.user,
            date_published_end=timezone.now() + timedelta(days=2),
            text="post text",
            image_url="/path/to/image",
            publisher=cls.publisher,
        )
        cls.related_post, __ = Post.objects.get_or_create(
            title="related title",
            og_title="related og title",
            twitter_title="related twitter title",
            schemaorg_title="related schemaorg title",
            schemaorg_description="related schemaorg description",
            slug="related-title",
            abstract="related post abstract",
            meta_description="related post meta",
            meta_keywords="related post keyword1,related post keyword 2",
            author=cls.user,
            date_published_end=timezone.now() + timedelta(days=2),
            text="related post text",
            image_url="/path/to/related-image",
        )
        cls.comment, __ = Comment.objects.get_or_create(
            body="comment body",
            post=cls.post,
        )
        cls.post.related_posts.add(cls.related_post)
        cls.post.main_image, __ = cls.create_django_image()
        cls.post.save()
        cls.related_post.main_image, __ = cls.create_django_image()
        cls.related_post.save()
        cls.image_url = cls.post.main_image.url
        cls.related_image_url = cls.related_post.main_image.url
        cls.image_width = cls.post.main_image.width
        cls.image_height = cls.post.main_image.height

    @override_settings(META_SITE_PROTOCOL="http", META_USE_SITES=True)
    def test_as_meta(self):
        expected = {
            "locale": "dummy_locale",
            "image": "http://example.com{}".format(self.image_url),
            "image_width": self.image_width,
            "image_height": self.image_height,
            "object_type": "Article",
            "tag": None,
            "keywords": ["post keyword1", "post keyword 2"],
            "og_profile_id": "1111111111111",
            "twitter_description": "post meta",
            "schemaorg_type": "Article",
            "title": "a title",
            "og_title": "og title",
            "twitter_title": "twitter title",
            "schemaorg_title": "schemaorg title",
            "schemaorg_description": "schemaorg description",
            "expiration_time": self.post.date_published_end,
            "og_description": "post meta",
            "description": "post meta",
            "twitter_type": "Summary",
            "modified_time": self.post.date_modified,
            "og_author_url": "https://facebook.com/foo.bar",
            "og_app_id": "appid",
            "fb_pages": "fbpages",
            "published_time": self.post.date_published,
            "url": "http://example.com/title/",
            "og_publisher": "https://facebook.com/foo.blag",
            "og_type": "Article",
            "twitter_author": "@FooBar",
            "twitter_site": "@FooBlag",
            "false_prop": False,
            "other_prop": "get_other_prop",
            "custom_namespace": ["foo", "bar"],
            "extra_props": {"key": "val"},
            "extra_custom_props": [
                ("custom1", "custom_name1", "custom_val1"),
                ("custom2", "custom_name2", "custom_val2"),
            ],
            "schema": {
                "@type": "Article",
                "image": "http://example.com{}".format(self.image_url),
                "articleBody": "post text",
                "articleSection": ["category 1", "category 2"],
                "author": {
                    "@type": "Person",
                    "name": self.post.author.get_full_name(),
                },
                "copyrightYear": self.post.date_published.year,
                "dateCreated": self.post.date_created.isoformat(),
                "dateModified": self.post.date_modified.isoformat(),
                "datePublished": self.post.date_published.isoformat(),
                "expires": self.post.date_published_end.isoformat(),
                "headline": "post abstract",
                "keywords": ["post keyword1", "post keyword 2"],
                "description": "post meta",
                "name": "a title",
                "url": "http://example.com/title/",
                "mainEntityOfPage": {
                    "@type": "WebPage",
                    "@id": "http://example.com/title/",
                },
                "publisher": {
                    "@type": "Organization",
                    "logo": {
                        "@type": "ImageObject",
                        "url": "http://example.com/some/logo.png",
                    },
                    "name": "publisher name",
                },
                "comment": [{"@type": "Comment", "text": "comment body"}],
                "commentCount": self.post.comments.count(),
                "citation": [
                    {
                        "@type": "Article",
                        "articleBody": "related post text",
                        "articleSection": ["category 1", "category 2"],
                        "author": {"@type": "Person", "name": self.related_post.author.get_full_name()},
                        "citation": [],
                        "comment": [],
                        "commentCount": self.related_post.comments.count(),
                        "copyrightYear": self.related_post.date_published.year,
                        "dateCreated": self.related_post.date_created.isoformat(),
                        "dateModified": self.related_post.date_modified.isoformat(),
                        "datePublished": self.related_post.date_published.isoformat(),
                        "description": "related post meta",
                        "expires": self.related_post.date_published_end.isoformat(),
                        "headline": "related post abstract",
                        "image": "http://example.com{}".format(self.related_image_url),
                        "keywords": ["related post keyword1", "related post keyword 2"],
                        "mainEntityOfPage": {"@id": "http://example.com/related-title/", "@type": "WebPage"},
                        "name": "related title",
                        "publisher": None,
                        "url": "http://example.com/related-title/",
                    }
                ],
            },
        }
        with override_settings(META_OG_NAMESPACES=["foo", "bar"], META_FB_PAGES="fbpages", META_FB_APPID="appid"):
            meta = self.post.as_meta()
            self.assertTrue(meta)
            for key in expected.keys():
                value = expected[key]
                if value is not None:
                    self.assertEqual(value, getattr(meta, key))
                else:
                    self.assertFalse(hasattr(meta, key))

            self.assertEqual(self.post.build_absolute_uri("hi"), "http://example.com/hi")
            self.assertEqual(self.post.build_absolute_uri("http://example.com/hi"), "http://example.com/hi")

    @override_settings(META_SITE_PROTOCOL="http", META_FB_PAGES="fbpages", META_FB_APPID="appid")
    def test_as_meta_with_request(self):
        # Server is different as it's taken directly from the request object
        media = {
            "url": "https://testserver{}".format(self.image_url),
            "width": self.image_width,
            "height": self.image_height,
            "alt": "a title",
        }
        expected = {
            "locale": "dummy_locale",
            "image_object": media,
            "image": "https://testserver{}".format(self.image_url),
            "image_width": self.image_width,
            "image_height": self.image_height,
            "object_type": "Article",
            "tag": False,
            "keywords": ["post keyword1", "post keyword 2"],
            "og_profile_id": "1111111111111",
            "twitter_description": "post meta",
            "schemaorg_type": "Article",
            "title": "a title",
            "og_title": "og title",
            "twitter_title": "twitter title",
            "schemaorg_title": "schemaorg title",
            "schemaorg_description": "schemaorg description",
            "expiration_time": self.post.date_published_end,
            "og_description": "post meta",
            "description": "post meta",
            "twitter_type": "Summary",
            "modified_time": self.post.date_modified,
            "og_author_url": "https://facebook.com/foo.bar",
            "og_app_id": "appid",
            "fb_pages": "fbpages",
            "published_time": self.post.date_published,
            "url": "https://testserver/title/",
            "og_publisher": "https://facebook.com/foo.blag",
            "og_type": "Article",
            "twitter_author": "@FooBar",
            "twitter_site": "@FooBlag",
            "custom_namespace": None,
            "extra_props": {"key": "val"},
            "extra_custom_props": [
                ("custom1", "custom_name1", "custom_val1"),
                ("custom2", "custom_name2", "custom_val2"),
            ],
            "schema": {
                "@type": "Article",
                "image": "http://example.com{}".format(self.image_url),
                "articleBody": "post text",
                "articleSection": ["category 1", "category 2"],
                "author": {
                    "@type": "Person",
                    "name": self.post.author.get_full_name(),
                },
                "copyrightYear": self.post.date_published.year,
                "dateCreated": self.post.date_created.isoformat(),
                "dateModified": self.post.date_modified.isoformat(),
                "datePublished": self.post.date_published.isoformat(),
                "expires": self.post.date_published_end.isoformat(),
                "headline": "post abstract",
                "keywords": ["post keyword1", "post keyword 2"],
                "description": "post meta",
                "name": "a title",
                "url": "http://example.com/title/",
                "mainEntityOfPage": {
                    "@type": "WebPage",
                    "@id": "http://example.com/title/",
                },
                "publisher": {
                    "@type": "Organization",
                    "logo": {
                        "@type": "ImageObject",
                        "url": "http://example.com/some/logo.png",
                    },
                    "name": "publisher name",
                },
                "comment": [{"@type": "Comment", "text": "comment body"}],
                "commentCount": self.post.comments.count(),
                "citation": [
                    {
                        "@type": "Article",
                        "articleBody": "related post text",
                        "articleSection": ["category 1", "category 2"],
                        "author": {"@type": "Person", "name": self.related_post.author.get_full_name()},
                        "citation": [],
                        "comment": [],
                        "commentCount": self.related_post.comments.count(),
                        "copyrightYear": self.related_post.date_published.year,
                        "dateCreated": self.related_post.date_created.isoformat(),
                        "dateModified": self.related_post.date_modified.isoformat(),
                        "datePublished": self.related_post.date_published.isoformat(),
                        "description": "related post meta",
                        "expires": self.related_post.date_published_end.isoformat(),
                        "headline": "related post abstract",
                        "image": "http://example.com{}".format(self.related_image_url),
                        "keywords": ["related post keyword1", "related post keyword 2"],
                        "mainEntityOfPage": {"@id": "http://example.com/related-title/", "@type": "WebPage"},
                        "name": "related title",
                        "publisher": None,
                        "url": "http://example.com/related-title/",
                    }
                ],
            },
        }
        request = self.get_request(None, "en", path="/title/", secure=True)
        meta = self.post.as_meta(request)
        self.assertTrue(meta)
        for key in ModelMeta._metadata_default.keys():
            value = expected[key]
            if value is not False:
                self.assertEqual(value, getattr(meta, key))
            else:
                self.assertFalse(hasattr(meta, key))

    @override_settings(META_SITE_PROTOCOL="http")
    def test_as_meta_get_request_deprecation(self):
        request = self.get_request(None, "en", path="/title/", secure=True)
        with warnings.catch_warnings(record=True) as w:
            warnings.simplefilter("always")
            self.post.get_meta(request)
            assert len(w) == 0

        with warnings.catch_warnings(record=True) as w:
            warnings.simplefilter("always")
            self.post.get_request()
            assert len(w) == 1
            assert issubclass(w[-1].category, PendingDeprecationWarning)

    @override_settings(
        META_USE_SITES=True,
        META_SITE_PROTOCOL="http",
        META_USE_OG_PROPERTIES=True,
        META_USE_TWITTER_PROPERTIES=True,
        META_USE_SCHEMAORG_PROPERTIES=True,
        META_USE_JSON_LD_SCHEMA=True,
    )
    def test_templatetag(self):
        self.post.as_meta()
        response = self.client.get("/title/")
        self.assertContains(response, '<html  itemscope itemtype="https://schema.org/Article" >')
        self.assertNotContains(response, '    itemscope itemtype="https://schema.org/Article"')
        self.assertContains(response, 'article:published_time"')
        self.assertContains(
            response, '<meta name="twitter:image" content="http://example.com{}">'.format(self.image_url)
        )
        self.assertContains(
            response, '<meta itemprop="description" content="{}">'.format(self.post.schemaorg_description)
        )
        self.assertContains(
            response, '<meta name="twitter:description" content="{}">'.format(self.post.meta_description)
        )
        self.assertContains(
            response, '<meta property="og:description" content="{}">'.format(self.post.meta_description)
        )
        self.assertContains(response, '<meta name="description" content="{}">'.format(self.post.meta_description))
        self.assertContains(
            response, '<meta name="keywords" content="{}">'.format(", ".join(self.post.meta_keywords.split(",")))
        )
        self.assertContains(response, '<meta name="key" content="val">')
        self.assertContains(response, '<meta custom1="custom_name1" content="custom_val1">')
        self.assertContains(response, '<meta custom2="custom_name2" content="custom_val2">')
        self.assertContains(response, '<meta property="og:{}:alt" content="a title">'.format("image"))
        self.assertContains(response, '<meta property="og:{}:height" content="{}">'.format("image", self.image_height))
        self.assertContains(response, '<meta property="og:{}:width" content="{}">'.format("image", self.image_width))
        self.assertContains(
            response,
            '<meta property="og:{}:url" content="{}'.format("image", "http://example.com{}".format(self.image_url)),
        )
        print(response.content)
        self.assertContains(
            response, '<script type="application/ld+json">{}</script>'.format(self.post.as_meta().as_json_ld())
        )

    @override_settings(
        META_SITE_PROTOCOL="http",
        META_USE_SITES=True,
        META_USE_OG_PROPERTIES=True,
        META_USE_TWITTER_PROPERTIES=True,
        META_USE_SCHEMAORG_PROPERTIES=True,
    )
    def test_templatetag_metadatamixin(self):
        """
        Test for issue #11
        """
        response = self.client.get("/mixin/title/")
        self.assertContains(response, '<meta itemprop="description" content="{}">'.format(self.post.meta_description))
        self.assertContains(
            response, '<meta name="twitter:description" content="{}">'.format(self.post.meta_description)
        )
        self.assertContains(
            response, '<meta property="og:description" content="{}">'.format(self.post.meta_description)
        )
        self.assertContains(response, '<meta name="description" content="{}">'.format(self.post.meta_description))
        self.assertContains(
            response, '<meta name="keywords" content="{}">'.format(", ".join(self.post.meta_keywords.split(",")))
        )
        self.assertContains(response, '<meta name="twitter:image" content="http://example.com/path/to/image">')

    @override_settings(
        META_USE_SITES=True,
        META_SITE_PROTOCOL="http",
        META_USE_OG_PROPERTIES=True,
        META_USE_TWITTER_PROPERTIES=True,
        META_USE_SCHEMAORG_PROPERTIES=True,
    )
    def test_templatetag_metadatamixin_image_object(self):
        """
        Test for issue #11
        """
        response = self.client.get("/mixin_image/title/")
        self.assertContains(response, '<meta itemprop="description" content="{}">'.format(self.post.meta_description))
        self.assertContains(
            response, '<meta name="twitter:description" content="{}">'.format(self.post.meta_description)
        )
        self.assertContains(
            response, '<meta property="og:description" content="{}">'.format(self.post.meta_description)
        )
        self.assertContains(response, '<meta name="description" content="{}">'.format(self.post.meta_description))
        self.assertContains(
            response, '<meta name="keywords" content="{}">'.format(", ".join(self.post.meta_keywords.split(",")))
        )
        self.assertContains(
            response, '<meta name="twitter:image" content="http://example.com{}">'.format(self.image_url)
        )
        self.assertContains(response, '<meta property="og:{}:alt" content="a title">'.format("image"))
        self.assertContains(response, '<meta property="og:{}:height" content="{}">'.format("image", self.image_height))
        self.assertContains(response, '<meta property="og:{}:width" content="{}">'.format("image", self.image_width))
        self.assertContains(
            response,
            '<meta property="og:{}:url" content="{}'.format("image", "http://example.com{}".format(self.image_url)),
        )

    @override_settings(META_SITE_DOMAIN="example.com", META_USE_OG_PROPERTIES=True)
    def test_templatetag_secure_image(self):
        """
        Test for issue #79
        """
        with override_settings(META_SITE_PROTOCOL="http"):
            response = self.client.get("/mixin/title/")
            self.assertContains(response, '<meta property="og:image" content="http://example.com/path/to/image">')
        with override_settings(META_SITE_PROTOCOL="https"):
            response = self.client.get("/mixin/title/")
            self.assertContains(response, '<meta property="og:image" content="https://example.com/path/to/image">')
            self.assertContains(
                response, '<meta property="og:image:secure_url" content="https://example.com/path/to/image">'
            )

    @override_settings(
        META_USE_SITES=True,
        META_USE_OG_PROPERTIES=False,
        META_USE_TWITTER_PROPERTIES=True,
        META_SITE_PROTOCOL="http",
        META_USE_SCHEMAORG_PROPERTIES=True,
    )
    def test_templatetag_no_og(self):
        response = self.client.get("/title/")
        self.assertFalse(response.rendered_content.find("og:description") > -1)
        self.assertContains(
            response, '<meta itemprop="description" content="{}">'.format(self.post.schemaorg_description)
        )
        self.assertContains(
            response, '<meta name="twitter:description" content="{}">'.format(self.post.meta_description)
        )
        self.assertContains(
            response, '<meta name="keywords" content="{}">'.format(", ".join(self.post.meta_keywords.split(",")))
        )

    @override_settings(META_OG_NAMESPACES=["foo", "bar"], META_USE_SITES=True, META_SITE_PROTOCOL="http")
    def test_templatetag_custom_namespaces(self):
        response = self.client.get("/title/")
        for ns in get_setting("OG_NAMESPACES"):
            self.assertContains(response, "{0}: http://ogp.me/ns/{0}#".format(ns))

    def test_generic_prop_basically_works(self):
        """
        Test vendorized generic_prop templatetag
        """
        self.assertEqual(generic_prop("og", "type", "website"), '<meta property="og:type" content="website">')

    def test_google_plus_scope_works(self):
        """
        Test vendorized googleplus_scope templatetag
        """
        self.assertEqual(googleplus_html_scope("bar"), ' itemscope itemtype="https://schema.org/bar" ')

    @override_settings(META_USE_SITES=True, META_SITE_PROTOCOL="https")
    def test_image_protocol_https(self):
        meta = self.post.as_meta()
        self.assertEqual("https://example.com{}".format(self.image_url), meta.image)

    @override_settings(META_USE_SITES=True, META_SITE_PROTOCOL="http")
    def test_image_protocol_http(self):
        meta = self.post.as_meta()
        self.assertEqual("http://example.com{}".format(self.image_url), meta.image)

    def test_not_use_sites(self):
        with override_settings(META_USE_SITES=False):
            with self.assertRaises(RuntimeError):
                self.post.as_meta()
        with override_settings(META_USE_SITES=True, META_SITE_PROTOCOL="http"):
            meta = self.post.as_meta()
            self.assertEqual(meta.url, "http://example.com/title/")

    @override_settings(META_SITE_PROTOCOL=None)
    def test_get_meta_protocol_without_site_protocol_will_raise(self):
        with self.assertRaises(ImproperlyConfigured):
            self.post.get_meta_protocol()

    def test_get_meta_protocol(self):
        with override_settings(META_SITE_PROTOCOL="http"):
            self.assertEqual(self.post.get_meta_protocol(), "http")
        with override_settings(META_SITE_PROTOCOL="https"):
            self.assertEqual(self.post.get_meta_protocol(), "https")

    def test_get_author_schemaorg(self):
        self.assertEqual(self.post.get_author_schemaorg(), "https://schemaorg-profile.com")



from django.conf import settings as django_settings
from django.utils.translation import gettext_lazy as _

META_SITE_PROTOCOL = None
META_SITE_DOMAIN = None
META_SITE_TYPE = None
META_SITE_NAME = None
META_INCLUDE_KEYWORDS = []
META_DEFAULT_KEYWORDS = []
META_IMAGE_URL = django_settings.STATIC_URL
META_USE_OG_PROPERTIES = False
META_USE_TWITTER_PROPERTIES = False
META_USE_FACEBOOK_PROPERTIES = False
META_USE_SCHEMAORG_PROPERTIES = False
META_USE_JSON_LD_SCHEMA = False
META_USE_SITES = False
META_USE_TITLE_TAG = False
META_OG_NAMESPACES = None


OBJECT_TYPES = (
    ("Article", _("Article")),
    ("Website", _("Website")),
)
TWITTER_TYPES = (
    ("summary", _("Summary Card")),
    ("summary_large_image", _("Summary Card with Large Image")),
    ("product", _("Product")),
    ("photo", _("Photo")),
    ("player", _("Player")),
    ("app", _("App")),
)
FB_TYPES = OBJECT_TYPES
SCHEMAORG_TYPES = (
    ("Article", _("Article")),
    ("Blog", _("Blog")),
    ("WebPage", _("Page")),
    ("WebSite", _("WebSite")),
    ("Event", _("Event")),
    ("Product", _("Product")),
    ("Place", _("Place")),
    ("Person", _("Person")),
    ("Book", _("Book")),
    ("LocalBusiness", _("LocalBusiness")),
    ("Organization", _("Organization")),
    ("Review", _("Review")),
)

META_OG_SECURE_URL_ITEMS = ("image", "audio", "video")
META_DEFAULT_IMAGE = ""
META_DEFAULT_TYPE = OBJECT_TYPES[0][0]
META_FB_TYPE = OBJECT_TYPES[0][0]
META_FB_TYPES = FB_TYPES
META_FB_APPID = ""
META_FB_PROFILE_ID = ""
META_FB_PUBLISHER = ""
META_FB_AUTHOR_URL = ""
META_FB_PAGES = ""
META_TWITTER_TYPE = TWITTER_TYPES[0][0]
META_TWITTER_TYPES = TWITTER_TYPES
META_TWITTER_SITE = ""
META_TWITTER_AUTHOR = ""
META_SCHEMAORG_TYPE = SCHEMAORG_TYPES[0][0]
META_SCHEMAORG_TYPES = SCHEMAORG_TYPES

params = {param: value for param, value in locals().items() if param.startswith("META_")}


def get_setting(name):
    """Get setting value from django settings with fallback to globals defaults."""
    from django.conf import settings

    return getattr(settings, "META_%s" % name, params["META_%s" % name])



import json
import warnings
from datetime import date

from django.apps import apps
from django.core.exceptions import ImproperlyConfigured

from .settings import get_setting

# This global variable is used to cache schemas for visited models to avoid recursion errors when traversing
# when parent and its children has a reference to each other
# When one objects if visited, its schema is put there, in a key generated from its pk
# By adding here visited items, it can be used as a local cache, to stop recursion
visited = {}


class FullUrlMixin:
    """
    Provides a few convenience methods to retrieve the full URL (which includes protocol and domain) of an object.

    If possible, :py:meth:`django.http.request.HttpRequest.build_absolute_uri` is used
    """

    def get_domain(self):
        """
        Discover the current website domain

        :py:class:`django.contrib.sites.models.Site`
        and :ref:`META_SITE_DOMAIN`
        (in this order) are used

        :return: domain URL
        """
        try:
            use_site = self.use_sites
        except AttributeError:
            use_site = get_setting("USE_SITES")

        if use_site:
            try:
                Site = apps.get_model("sites.Site")
                try:
                    return Site.objects.get_current(self.request).domain
                except AttributeError:
                    return Site.objects.get_current().domain
            except LookupError:
                raise ImproperlyConfigured("Add django.contrib.sites to INSTALLED_APPS because META_USE_SITES is True")
        if not get_setting("SITE_DOMAIN"):
            raise ImproperlyConfigured("META_SITE_DOMAIN is not set")
        return get_setting("SITE_DOMAIN")

    def get_protocol(self):
        """
        Discover the current website protocol from :ref:`META_SITE_PROTOCOL`

        :return: http or https depending on :ref:`META_SITE_PROTOCOL`
        """
        if not get_setting("SITE_PROTOCOL"):
            raise ImproperlyConfigured("META_SITE_PROTOCOL is not set")
        return get_setting("SITE_PROTOCOL")

    def _get_full_url(self, url):
        """
        Build the full URL (protocol and domain included) for the URL given as argument

        :param url: absolute (domain-less) URL
        :return: full url
        """
        try:
            return self.request.build_absolute_uri(url)
        except AttributeError:
            pass
        if not url:
            return None
        if url.startswith("http"):
            return url
        meta_protocol = self.get_protocol()
        domain = self.get_domain()
        separator = "://"
        if url.startswith("//"):
            separator = ":"
            domain = ""
        elif not url.startswith("/"):
            url = "/%s" % url
        if domain.startswith("http"):
            meta_protocol = ""
            separator = ""
        return "{meta_protocol}{separator}{domain}{url}".format(
            meta_protocol=meta_protocol, separator=separator, domain=domain, url=url
        )


class Meta(FullUrlMixin):
    """
    Helper for building context meta object
    """

    _keywords = []
    _url = None
    _image = None
    _image_object = None
    _schema = {}
    """
    Base schema.org types definition.

    It's a dictionary containing all the schema.org properties for the described objects.

    See :ref:`a sample implementation <schema._schema>`.
    """
    request = None
    _obj = None
    """
    Linked :py:class:`~meta.models.ModelMeta` instance (if Meta is generated from a ModelMeta object)
    """

    def __init__(self, **kwargs):
        self.request = kwargs.get("request", None)
        self.use_sites = kwargs.get("use_sites", get_setting("USE_SITES"))
        self.title = kwargs.get("title")
        self.og_title = kwargs.get("og_title")
        self.twitter_title = kwargs.get("twitter_title")
        self.schemaorg_title = kwargs.get("schemaorg_title")
        self.schemaorg_description = kwargs.get("schemaorg_description")
        self.description = kwargs.get("description")
        self.extra_props = kwargs.get("extra_props")
        self.extra_custom_props = kwargs.get("extra_custom_props")
        self.custom_namespace = kwargs.get("custom_namespace", get_setting("OG_NAMESPACES"))
        self.keywords = kwargs.get("keywords")
        self.url = kwargs.get("url")
        self.image = kwargs.get("image")
        self.image_object = kwargs.get("image_object")
        self.image_width = kwargs.get("image_width")
        self.image_height = kwargs.get("image_height")
        self.object_type = kwargs.get("object_type", get_setting("SITE_TYPE"))
        self.site_name = kwargs.get("site_name", get_setting("SITE_NAME"))
        self.twitter_site = kwargs.get("twitter_site")
        self.twitter_creator = kwargs.get("twitter_creator")
        self.twitter_type = kwargs.get("twitter_type", kwargs.get("twitter_card", get_setting("TWITTER_TYPE")))
        self.twitter_card = self.twitter_type
        self.facebook_app_id = kwargs.get("facebook_app_id")
        self.locale = kwargs.get("locale")
        self.use_og = kwargs.get("use_og", get_setting("USE_OG_PROPERTIES"))
        self.use_twitter = kwargs.get("use_twitter", get_setting("USE_TWITTER_PROPERTIES"))
        self.use_facebook = kwargs.get("use_facebook", get_setting("USE_FACEBOOK_PROPERTIES"))
        self.use_schemaorg = kwargs.get("use_schemaorg", get_setting("USE_SCHEMAORG_PROPERTIES"))
        self.use_json_ld = kwargs.get("use_json_ld", get_setting("USE_JSON_LD_SCHEMA"))
        self.use_title_tag = kwargs.get("use_title_tag", get_setting("USE_TITLE_TAG"))
        self.schemaorg_type = kwargs.get("schemaorg_type", get_setting("SCHEMAORG_TYPE"))
        self.fb_pages = kwargs.get("fb_pages", get_setting("FB_PAGES"))
        self.og_app_id = kwargs.get("og_app_id", get_setting("FB_APPID"))
        self._schema = kwargs.get("schema", {})
        self._obj = kwargs.get("obj", {})

    @property
    def keywords(self):
        return self._keywords

    @keywords.setter
    def keywords(self, keywords):
        if keywords is None:
            kws = get_setting("DEFAULT_KEYWORDS")
        else:
            if not hasattr(keywords, "__iter__"):
                # Not iterable
                raise ValueError("Keywords must be an iterable")
            kws = list(keywords)
            if get_setting("INCLUDE_KEYWORDS"):
                kws += get_setting("INCLUDE_KEYWORDS")
        seen = set()
        seen_add = seen.add
        self._keywords = [k for k in kws if k not in seen and not seen_add(k)]

    def get_full_url(self, url):
        return self._get_full_url(url)

    @property
    def url(self):
        return self._url

    @url.setter
    def url(self, url):
        self._url = self._get_full_url(url)

    def _normalize_media_url(self, url):
        if not url.startswith("http") and not url.startswith("/"):
            url = "{}{}".format(get_setting("IMAGE_URL"), url)
        return self.get_full_url(url)

    @property
    def image(self):
        if self.image_object:
            return self.image_object.get("url")
        return self._image

    @image.setter
    def image(self, image):
        if image is None and get_setting("DEFAULT_IMAGE"):
            image = get_setting("DEFAULT_IMAGE")
        if image:
            self._image = self._normalize_media_url(image)

    @property
    def image_object(self):
        return self._image_object

    @image_object.setter
    def image_object(self, image):
        try:
            if image:
                image["url"] = self._normalize_media_url(image.get("url", None))
                if self.get_protocol() == "https":
                    secure_fallback_url = image.get("secure_url", image.get("url", None))
                    image["secure_url"] = self._normalize_media_url(secure_fallback_url)
                    if image["secure_url"].startswith("http://"):
                        image["secure_url"] = image["secure_url"].replace("http://", "https://")
                self._image_object = image
        except KeyError:
            self._image_object = None

    @property
    def schema(self):
        """
        Schema.org object description.

        Items in the schema are converted in a format suitable of json encoding at this stage:

        * instances of :py:class:`~meta.views.Meta` as their schema
        * dates as isoformat
        * iterables and dicts are processed depth-first to process their items

        If no type is set :py:attr:`~meta.views.Meta.schemaorg_type` is used

        :return: dict
        """
        from meta.models import ModelMeta

        def process_item(item):
            if isinstance(item, Meta):
                return item.schema
            if isinstance(item, ModelMeta):
                # if not cached, object schema is generated and put into local cache
                if item._local_key not in visited:
                    visited[item._local_key] = item.as_meta(self.request).schema
                return visited[item._local_key]
            elif isinstance(item, date):
                return item.isoformat()
            elif isinstance(item, list) or isinstance(item, tuple):
                return [process_item(value) for value in item]
            elif isinstance(item, dict):
                return {itemkey: process_item(itemvalue) for itemkey, itemvalue in item.items()}
            return item

        schema = {}
        # object is immediately set here to recursion
        # if we are visiting parent -> child relation, we don't need the pointer
        # back up
        if isinstance(self._obj, ModelMeta):
            visited[self._obj._local_key] = None
        for key, val in self._schema.items():
            schema[key] = process_item(val)
        if "@type" not in schema:
            schema["@type"] = self.schemaorg_type
        # after generating the full schema, we can save it in the local cache for future uses
        if isinstance(self._obj, ModelMeta):
            visited[self._obj._local_key] = schema
        return schema

    @schema.setter
    def schema(self, schema):
        self._schema = schema

    def as_json_ld(self):
        """
        Convert the schema to json-ld

        :return: json
        """
        data = self.schema
        data["@context"] = "http://schema.org"
        return json.dumps(data)


class MetadataMixin(FullUrlMixin):
    """
    Django CBV mixin to prepare metadata for the view context
    """

    meta_class = Meta
    context_meta_name = "meta"

    title = None
    og_title = None
    twitter_title = None
    schemaorg_title = None
    schemaorg_description = None
    description = None
    extra_props = None
    extra_custom_props = None
    custom_namespace = None
    keywords = []
    url = None
    image = None
    image_object = None
    object_type = None
    site_name = None
    twitter_site = None
    twitter_creator = None
    twitter_type = None
    facebook_app_id = None
    locale = None
    use_sites = False
    use_og = False
    use_title_tag = False
    schemaorg_type = None
    schema = {}

    def __init__(self, **kwargs):
        self.use_sites = get_setting("USE_SITES")
        self.use_og = get_setting("USE_OG_PROPERTIES")
        self.use_title_tag = get_setting("USE_TITLE_TAG")
        super().__init__(**kwargs)

    def get_meta_class(self):
        return self.meta_class

    def get_protocol(self):
        return get_setting("SITE_PROTOCOL")

    def get_domain(self):
        return get_setting("SITE_DOMAIN")

    def get_meta_title(self, context=None):
        return self.title

    def get_meta_og_title(self, context=None):
        return self.og_title

    def get_meta_twitter_title(self, context=None):
        return self.twitter_title

    def get_meta_schemaorg_title(self, context=None):
        return self.schemaorg_title

    def get_meta_schemaorg_description(self, context=None):
        return self.schemaorg_description

    def get_meta_description(self, context=None):
        return self.description

    def get_meta_keywords(self, context=None):
        return self.keywords

    def get_meta_url(self, context=None):
        return self.url

    def get_meta_image(self, context=None):
        if self.image_object and self.image_object.get("url", None):
            return self.image_object["url"]
        return self.image

    def get_meta_image_object(self, context=None):
        return self.image_object

    def get_meta_object_type(self, context=None):
        return self.object_type or get_setting("SITE_TYPE")

    def get_meta_site_name(self, context=None):
        return self.site_name or get_setting("SITE_NAME")

    def get_meta_extra_props(self, context=None):
        return self.extra_props

    def get_meta_extra_custom_props(self, context=None):
        return self.extra_custom_props

    def get_meta_custom_namespace(self, context=None):
        return self.custom_namespace or get_setting("OG_NAMESPACES")

    def get_meta_twitter_site(self, context=None):
        return self.twitter_site

    def get_meta_twitter_creator(self, context=None):
        return self.twitter_creator

    @property
    def twitter_card(self):
        warnings.warn("twitter_card attribute will be removed in version 3.0", PendingDeprecationWarning, stacklevel=2)
        return self.twitter_type

    @twitter_card.setter
    def twitter_card(self, value):
        warnings.warn("twitter_card attribute will be removed in version 3.0", PendingDeprecationWarning, stacklevel=2)
        self.twitter_type = value

    def get_meta_twitter_card(self, context=None):
        warnings.warn(
            "get_meta_twitter_card attribute will be removed in version 3.0", PendingDeprecationWarning, stacklevel=2
        )
        return self.twitter_type

    def get_meta_twitter_type(self, context=None):
        return self.twitter_type

    def get_meta_facebook_app_id(self, context=None):
        return self.facebook_app_id

    def get_meta_schemaorg_type(self, context=None):
        return self.schemaorg_type

    def get_meta_locale(self, context=None):
        return self.locale

    def get_schema(self, context=None):
        """
        The generic API to retrieve the full schema.org structure for the view.

        By default it returns the :py:attr:`schema`. You can reimplement this method
        to build the schema.org structure at runtime. See :ref:`a sample implementation <schema.get_schema>`.

        :param context: view context
        :return: dictionary
        """
        return self.schema

    def get_schema_property(self, schema_type, property, context=None):
        """
        The generic API to retrieve the attribute value for a generic schema type

        This is just a stub that **must** be implemented

        :param schema_type: name of the schema type
        :param property: name of the property
        :param context: view context
        :return: property value
        """
        raise NotImplementedError

    def get_meta(self, context=None):
        return self.get_meta_class()(
            use_og=self.use_og,
            use_title_tag=self.use_title_tag,
            use_sites=self.use_sites,
            title=self.get_meta_title(context=context),
            og_title=self.get_meta_og_title(context=context),
            twitter_title=self.get_meta_twitter_title(context=context),
            schemaorg_title=self.get_meta_schemaorg_title(context=context),
            schemaorg_description=self.get_meta_schemaorg_description(context=context),
            description=self.get_meta_description(context=context),
            extra_props=self.get_meta_extra_props(context=context),
            extra_custom_props=self.get_meta_extra_custom_props(context=context),
            custom_namespace=self.get_meta_custom_namespace(context=context),
            keywords=self.get_meta_keywords(context=context),
            image=self.get_meta_image(context=context),
            image_object=self.get_meta_image_object(context=context),
            url=self.get_meta_url(context=context),
            object_type=self.get_meta_object_type(context=context),
            site_name=self.get_meta_site_name(context=context),
            twitter_site=self.get_meta_twitter_site(context=context),
            twitter_creator=self.get_meta_twitter_creator(context=context),
            twitter_type=self.get_meta_twitter_type(context=context),
            locale=self.get_meta_locale(context=context),
            facebook_app_id=self.get_meta_facebook_app_id(context=context),
            schemaorg_type=self.get_meta_schemaorg_type(context=context),
            schema=self.get_schema(context=context),
        )

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        context[self.context_meta_name] = self.get_meta(context=context)
        return context
